#!/usr/bin/python
# -*- encoding: utf-8; py-indent-offset: 4 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2018-03-17
# #
# rewrite of the original checkpoint_connections check from check_mk
#
# added connection statistic details for tcp/udp/icmp/other .....
#
# 2018-05-29: added connection limit (fwConnTableLimit)
# 2020-05-31: changed form checkpoint_connections to checkpoint_fw_connections
#             changed to checkpoint_fw_connections_default_levels to
#             factory_settings['checkpoint_fw_connections_defaults']
# 2020-06-07: code cleanup, added wato, added warn/crit for connection rate
# 2020-06-08: changed snmp-scan function, code cleanup
# 2021-08-24: rewritten for CMK 2.0,
#             added relative thresholds (idea and code based on cmk PR #312 by https://github.com/gradecke)
#             added lower levels and admin_table_limit
#
#
# sample info
# [[[u'559684419', u'203840211', u'51093794', u'786231', u'815404655', u'0']], [[u'11172', u'27598', u'0']]]
#
# no firewall
# [[], []]
#

import time
from dataclasses import dataclass
from typing import List, Dict, Any, Optional

from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    register,
    Service,
    Result,
    check_levels,
    State,
    SNMPTree,
    all_of,
    startswith,
    any_of,
    equals,
    Metric,
    get_rate,
    GetRateError,
    get_value_store,
    IgnoreResultsError,
    render,
)
from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import (
    DiscoveryResult,
    CheckResult,
    StringTable,
)


@dataclass
class CheckpointFwConnections:
    fwConnectionsTcp: int
    fwConnectionsUdp: int
    fwConnectionsIcmp: int
    fwConnectionsOther: int
    fwConnectionsSum: int
    fwConnectionRate: int
    fwCurrnetNumConn: int
    fwPeakNumConn: int
    fwConnTableLimit: int


def parse_checkpoint_fw_connections(string_table: List[StringTable]) -> Optional[CheckpointFwConnections]:
    fwConnectionsStat, fwpolicystat = string_table
    try:
        fwConnectionsTcp, fwConnectionsUdp, fwConnectionsIcmp, fwConnectionsOther, fwConnectionsSum, \
        fwConnectionRate = fwConnectionsStat[0]
    except(IndexError, ValueError):
        return

    fwCurrnetNumConn, fwPeakNumConn, fwConnTableLimit = fwpolicystat[0]

    return CheckpointFwConnections(
        fwConnectionsTcp=int(fwConnectionsTcp),
        fwConnectionsUdp=int(fwConnectionsUdp),
        fwConnectionsIcmp=int(fwConnectionsIcmp),
        fwConnectionsOther=int(fwConnectionsOther),
        fwConnectionsSum=int(fwConnectionsSum),
        fwConnectionRate=int(fwConnectionRate),
        fwCurrnetNumConn=int(fwCurrnetNumConn),
        fwPeakNumConn=int(fwPeakNumConn),
        fwConnTableLimit=int(fwConnTableLimit),
    )


def discovery_checkpoint_fw_connections(section: CheckpointFwConnections) -> DiscoveryResult:
    yield Service()


def check_checkpoint_fw_connections(params, section: CheckpointFwConnections) -> CheckResult:
    fwConnTableLimit = params.get('admin_table_limit', section.fwConnTableLimit)

    if fwConnTableLimit > 0:
        yield from check_levels(
            value=section.fwCurrnetNumConn * 100 / fwConnTableLimit,
            levels_upper=params.get('levels_upper_relative'),
            levels_lower=params.get('levels_lower_relative'),
            boundaries=(0, 100),
            label='Connections relative',
            render_func=render.percent,
            metric_name='checkpoint_fw_connections_relative'
        )

    if fwConnTableLimit > 0 and 'levels_upper_relative' in params:
        warn_pct, crit_pct = params['levels_upper_relative']
        warn = fwConnTableLimit * warn_pct / 100
        crit = fwConnTableLimit * crit_pct / 100
        levels_upper = (warn, crit)
    else:
        # use absolute levels if no relative levels provided or no maximum set on CP
        levels_upper = params.get('levels_upper_absolute', (None, None))

    if section.fwConnTableLimit > 0 and 'levels_lower_relative' in params:
        warn_pct, crit_pct = params['levels_lower_relative']
        warn = fwConnTableLimit * warn_pct / 100
        crit = fwConnTableLimit * crit_pct / 100
        levels_lower = (warn, crit)
    else:
        # use absolute levels if no relative levels provided or no maximum set on CP
        levels_lower = params.get('levels_lower_absolute', (None, None))

    for label, unit, metric, value, levels_upper, levels_lower in [
        ('Current', '', 'fwcurrentnumconn', section.fwCurrnetNumConn, levels_upper, levels_lower),
        ('Peak', '', 'fwpeaknumconn', section.fwPeakNumConn, (None, None), (None, None)),
        ('Rate', '/s', 'fwconnectionrate', section.fwConnectionRate, (None, None), (None, None)),
    ]:
        yield from check_levels(
            value=value,
            label=label,
            metric_name=f'checkpoint_fw_connections_{metric}',
            render_func=lambda v: f'{v:.0f}{unit}',
            levels_upper=levels_upper,
            levels_lower=levels_lower,
            boundaries=(0, None),
        )

    if fwConnTableLimit == 0:
        yield Result(state=State.OK, summary=f'Table limit: automatically adjusted')
    elif section.fwConnTableLimit == fwConnTableLimit:
        yield Result(state=State.OK, summary=f'Table limit: {section.fwConnTableLimit}')
    else:
        yield Result(state=State.OK, summary=f'Table limit: {section.fwConnTableLimit} (CMK admin limit')

    now_time = time.time()
    value_store = get_value_store()

    for key, value in [
        ('fwconnectionstcp', section.fwConnectionsTcp),
        ('fwconnectionsudp', section.fwConnectionsUdp),
        ('fwconnectionsicmp', section.fwConnectionsIcmp),
        ('fwconnectionsother', section.fwConnectionsOther),
        ('fwconnectionssum', section.fwConnectionsSum),
    ]:
        try:
            value = get_rate(value_store, f'checkpoint_fw_connections_{key}', now_time, value, raise_overflow=True)
        except GetRateError:
            value = 0
        yield Metric(name=f'checkpoint_fw_connections_{key}', value=value, boundaries=(0, None))


register.snmp_section(
    name='checkpoint_fw_connections',
    parse_function=parse_checkpoint_fw_connections,
    fetch=[
        SNMPTree(
            base='.1.3.6.1.4.1.2620.1.1.26.11',  # CHECKPOINT-MIB::fwConnectionsStat
            oids=[
                '1',  # fwConnectionsStatConnectionsTcp
                '2',  # fwConnectionsStatConnectionsUdp
                '3',  # fwConnectionsStatConnectionsIcmp
                '4',  # fwConnectionsStatConnectionsOther
                '5',  # fwConnectionsStatConnections
                '6',  # fwConnectionsStatConnectionRate
            ]
        ),
        SNMPTree(
            base='.1.3.6.1.4.1.2620.1.1.25',  # CHECKPOINT-MIB::fwPolicyStat
            oids=[
                '3',  # fwNumConn
                '4',  # fwPeakNumConn
                '10',  # fwConnTableLimit
            ]
        ),
    ],
    detect=any_of(
        startswith('.1.3.6.1.2.1.1.2.0', '.1.3.6.1.4.1.2620'),
        all_of(
            equals('.1.3.6.1.2.1.1.2.0', '.1.3.6.1.4.1.8072.3.2.10'),
            equals('.1.3.6.1.4.1.2620.1.6.1.0', 'SVN Foundation'),
        )
    )
)

register.check_plugin(
    name='checkpoint_fw_connections',
    service_name='Firewall connections',
    discovery_function=discovery_checkpoint_fw_connections,
    check_function=check_checkpoint_fw_connections,
    check_ruleset_name='checkpoint_fw_connections',
    check_default_parameters={
       # 'levels_upper_absolute': (10000, 20000),
       # 'levels_upper_relative': (75, 85)
    }
)
